package TecIO::ASCII;

use Exporter;
use Carp;

our @ISA = qw/Exporter/;
our @EXPORT = ();
our @EXPORT_OK = qw(fromstr tostr);

use strict;
use warnings;

my $str = qr/\s*"(?:[^\\"]|\\.)*"\s*/;
my $real = qr/[+-]?(?:\d+(?:\.\d*)?|\.\d+)(?:[Ee][+-]?\d+)?/;

sub fromstr($){
my $s = shift;
$s =~ s/\A\s*"?|"?\s*\Z//g;
$s =~ s/\\(.)/$1/g;
$s
};

sub tostr($){
my $t = shift;
$t =~ s/(["\\\n])/\\$1/g;
qq("$t")
};

=head2 DESCRIPTION

This module is an alpha-quality way to deal with data files of Tecplot
in ASCII format. Not all such files can be correctly opened and saved by
this module (it was designed for working with simple flow simulations created
in tutorial program S<Flos 3D>  written in SPbSTU). One should write an XS to
bind to perl original C<TecIO> binary library for WinNN/*nix/MacOS
(L<http://downloads.tecplot.com/tecio>).

=head3 Constructor new

  $db = TecIO::ASCII->new( qw/TITLE MyData FILETYPE FULL VARIABLES/,
    [qw/x y V/], zones=>[{qw/T zone1 I 10 J 20 DATAPACKING BLOCK/,
      data=>[@data]}] )

Creates new TecIO::ASCII database from data in format as displayed
in the example. In fact, this method currently just records its parameters
into hash and returns blessed reference to it.

=cut

sub new{ my ($class, %h) = splice @_; bless \%h => $class };

sub clone{

=head3 Method clone

  $bd = $db->clone(zones=>[qw/z1 z2/], vars=>[qw/v1 v2/])
  
This method copies all data structures from a database to new one.
Optionally, one can select zone set by listing their names or numbers in
C<zones> clause, and variables set in C<vars> clause. All
other things are copied without change. Note that order of
zones and variables in new database will be like in the arrays referred
in the clauses, so be careful with coordinates.

=cut

  my ($odb,%pars,$ndb,@zones,@vars) = splice @_;
  @zones = exists $pars{zones} ?
           map{ /\A\+?\d+\Z/ ? $_ : $odb->zones($_) }@{$pars{zones}}
                               : ( 0 .. ($odb->zones-1) );
  @vars = exists $pars{vars} ?
           map{ /\A\+?\d+\Z/ ? $_ : $odb->vars($_) }@{$pars{vars}}
                               : ();
  @$ndb{qw/TITLE FILETYPE/} = @$odb{qw/TITLE FILETYPE/};
  @{$ndb->{VARIABLES}} = @{$odb->VARIABLES}[@vars];
  foreach my $z(@{$odb->{zones}}[@zones]){
      my ($k,$v,$nz);
      push @{$ndb->{zones}}, $nz = {%$z};
      while( ($k => $v) = each %$nz ){
        if($k eq 'data'){
           if(@vars){
             my @h;
             if($z->{DATAPACKING} eq 'POINT'){
               foreach $k( 0 .. (@$v/@{$odb->VARIABLES}-1) ){
                foreach( @vars ){
                  push @h, $v->[ $k*@{$odb->VARIABLES} + $_ ]
                }
               }
             }else{#DATAPACKING eq BLOCK
               foreach( @vars ){
                  $k = $_*@$v/@{$odb->VARIABLES};
                  push @h, $v->[ $k..($k + @{$odb->VARIABLES}) ]
               }
             };
             $v = \@h
           }else{$v = [@$v]};
           $z->{data} = $v
        }else{
           $z->{$k} = ref($v) eq 'ARRAY' ? [@$v] : ref($v) eq 'HASH' ? {%$v} : $v
        }
      }
  };
  bless $ndb => ref $odb
};

sub Read{

=head3 Constructor Read

  $db = TecIO::ASCII->Read('data.tec')

=cut

my ($class,$fn,$self) = (shift,shift,{});
local (*F, $_);
{
  local $/ = undef;
  open F,'<',$fn or return;
  $_ = <F>   ;
  close F;
};
#comments
s/^\#.*?$//mg;
#header
/\b(?:TITLE\s*=($str))?(?:FILETYPE\s*=\s*(FULL|GRID|SOLUTION)\s*)?(?:VARIABLES\s*=((?:$str,?\s*)*))/i;
$self->{TITLE} = fromstr $1 if defined $1;
$self->{FILETYPE} = $2 if defined $2;     #/;
$self->{VARIABLES} = [];
if(defined $3){
  my $vs = $3;
  while($vs=~/($str)/g){
    push @{$self->{VARIABLES}}, fromstr($1)
  }
};
pos = 0;
READING:while(/\G(\b\w+\b|$str|$real|\(.*\)|\[.*\]|[=,])\s*/g){
  if(uc $1 eq 'ZONE'){
    #zone header
    push @{$self->{zones}}, my $z = { qw(ZONETYPE ORDERED DATAPACKING BLOCK) };
    while(/\G(AUXDATA\s+)?([A-Z]+)\s*=\s*(\w+|$str|$real|\(.*\)|\[.*\])\s*,?\s*/igc){
       if(defined $1){
         $z->{AUXDATA}{$2} = $3
       }elsif(uc $2 eq 'T'){
         $z->{T} = fromstr $3
       }else{
         $z->{uc $2 eq 'F' ? 'DATAPACKING' : uc $2} = $3;
       }
    };
BEGIN{no warnings;<<'COMMENT'};   #way of packing in multidim array
    #zone data - always stored in BLOCK packing!
    if($z->{DATAPACKING} eq 'BLOCK'){
      foreach my $varN( 0 .. $#{ $self->{VARIABLES} } ){
        if(exists $z->{I}){
          foreach my $I(1..$z->{I}){
            if(exists $z->{J}){
              foreach my $J(1..$z->{J}){
                if(exists $z->{K}){
                  foreach my $K(1..$z->{K}){
                    /\G(\S+)\s*/g or last READING; $z->{data}[$varN][$I-1][$J-1][$K-1] = $1;
                  }#foreach K
                }else{#K
                  /\G(\S+)\s*/g or last READING; $z->{data}[$varN][$I-1][$J-1] = $1
                }#if K
              }#foreach J
            }else{#J
              /\G(\S+)\s*/g or last READING; $z->{data}[$varN][$I-1] = $1
            }#if J
          }#foreach I
        }#if I
      };#foreach var
    }elsif($z->{DATAPACKING} eq 'POINT'){
      if(exists $z->{I}){
        foreach my $I(1..$z->{I}){
          if(exists $z->{J}){
            foreach my $J(1..$z->{J}){
              if(exists $z->{K}){
                foreach my $K(1..$z->{K}){
                  foreach my $varN( 0 .. $#{ $self->{VARIABLES} } ){
                    /\G(\S+)\s*/g or last READING; $z->{data}[$varN][$I-1][$J-1][$K-1] = $1
                  }#foreach var
                }#foreach K
              }else{#K
                foreach my $varN( 0 .. $#{ $self->{VARIABLES} } ){
                  /\G(\S+)\s*/g or last READING; $z->{data}[$varN][$I-1][$J-1] = $1
                }#foreach var
              }#if K
            }#foreach J
          }else{#J
            foreach my $varN( 0 .. $#{ $self->{VARIABLES} } ){
              /\G(\S+)\s*/g or last READING; $z->{data}[$varN][$I-1] = $1
            }#foreach var
          }#if J
        }#foreach I
      }#if I
    }; ### print "read zone $$z{T}\n"
COMMENT
    # Zone is stored in single array
    for my $n( 1 .. ($z->{I}//1)*($z->{J}//1)*($z->{K}//1)*@{$self->{VARIABLES}} ){   #/)){
       /\G(\S+)\s*/g;
       push @{$z->{data}}, $1
    }
  }#zone reading

}#READING;
bless $self => $class
};

=head3 C<AUTOLOAD>'ing

  $title = $db->TITLE;
  @{$db->VARIABLES}[3] = 'New varname'

C<AUTOLOAD> implements lvalue access to any component of database object
which is innerly stored as hash. Thus, C<$db->TITLE> means the same as
C<$db->{'TITLE'}> (and means title of the database; but methods C<TITLE>,
C<FILETYPE> and C<VARIABLES> will keep their semantics in several possible
future versions of the module, even if the data structure will evolve).

=cut
our $AUTOLOAD;
sub AUTOLOAD :lvalue{
  my $self = shift;
  $AUTOLOAD =~ /::([^:]+)\Z/;
  $self->{$1}
};

=head3 Methods

=over

=item S<C<String> --> represent whole database as string

  $ascii = $db->String()

=cut

sub String{
my $self = shift;
my $res = "TITLE = " . tostr($self->{TITLE}) .
  (exists $self->{FILETYPE} ?"\nFILETYPE = "  . $self->{FILETYPE} :'').
  "\nVARIABLES = " . join(', ',map{tostr $_}@{$self->{VARIABLES}});
foreach my $z(@{$self->{zones}}){
  my (@zop, $k, $v) = defined($z->{T}) ? 'T = '.tostr($z->{T}) : ();
  while( ($k=>$v) = each %$z ){
    push @zop, ( $k ne 'AUXDATA' ? "$k = $v"
                                 : join ', ', map{"AUXDATA $_ = $$v{$_}"}(keys %$v) )
            if $k ne 'data'  and  $k ne 'T'
  };
  $res .= "\nZONE " . join(', ', @zop);
  $k = '';
  foreach $v(@{$z->{data}}){
    if( length($k)+length($v) < 79 ){
      $k .= " $v"
    }else{
      $res .= "\n$k";
      $k = $v
    }
  };
  $res .= "\n$k\n"
};
return $res
};

sub Write{
  my ($self, $fn, $rew) = (shift, shift, shift);
  local *F;
  unless($rew){ return if -e $fn };
  open F, '>', $fn        or  return;
  binmode F;
  print F $self->String  or  return;
  close F                 or  return;
  return 1
};

=item S<C<getvar> --> get variable value

  $db->getvar( $var, $zone, $i?, $j?, $k?? )
  
=cut

sub getvar{
    my ($self,$var,$zone,$i,$j,$k) = splice @_;
    ($_//=1) -- foreach ($i, $j, $k);
    $zone = $self->zones($zone) // return  if  $zone !~ /\A\+?\d+/;
    $var = $self->vars($var) // return  if  $var !~ /\A\+?\d+/;
    $zone = $self->{zones}[$zone] // return;      #/;
    if($zone->{DATAPACKING} eq 'BLOCK'){
       return $zone->{data}[
        (
          ( $var*($zone->{K}//1) + $k )*($zone->{J}//1) + $j
        ) * ($zone->{I}//1) + $i                            #/  ))
       ]
    }elsif($zone->{DATAPACKING} eq 'POINT'){
       return $zone->{data}[
        ( (
          $k * ($zone->{J}//1) + $j
        ) * ($zone->{I}//1) + $i ) * $self->vars + $var  #/
       ]
    };
    carp 'Unknown data packing "',$zone->{DATAPACKING},'" in zone ',$zone->{'T'} if $^W;
    return
};

=item S<C<zones> --> get zone number by title or zone list/count

  $zone1no = $db->zones("Zone 1");
  foreach my $z( $db->zones ){ print $z->{T},"\n" };
  $zonecount = $db->zones();

=cut

sub zones{
  my($self,$tit) = (shift,shift);
  if(defined $tit){
    my $i = 0;
    foreach(@{$self->{zones}}){
      return $i if $_->{T} eq $tit
    }continue{$i++};
    return
  }else{
    return @{$self->{zones}}
  }
};

sub vars{
  my($self,$vid) = (shift,shift);
  if(defined $vid){
    my $i = 0;
    foreach(@{$self->{VARIABLES}}){
      return $i if $_ eq $vid
    }continue{$i++};
    return
  }else{
    return @{$self->{VARIABLES}}
  }
};

=back

=cut

1

